/* *********************************************************************** */ 
/*                           mus2sco v2.0                                  */
/*                       1990 Jean-Francois Giorgi                         */
/* *********************************************************************** */ 

/* ****************************************** */
/* readsmus : read IFF SMUS file              */
/* ****************************************** */
#include "smus2sco.h"

/* Message strings for IFFP codes. */ 
LOCAL char MsgOkay[]        = { "(IFF_OKAY) No FORM SMUS in the file." }; 
LOCAL char MsgEndMark[]     = { "(END_MARK) How did you get this message?" }; 
LOCAL char MsgDone[ ]       = { "(IFF_DONE) All done."}; 
LOCAL char MsgDos[]         = { "(DOS_ERROR) The DOS returned an error." }; 
LOCAL char MsgNot[]         = { "(NOT_IFF) Not an IFF file." }; 
LOCAL char MsgNoFile[]      = { "(NO_FILE) No such file found." }; 
LOCAL char MsgClientError[] = { "(CLIENT_ERROR) ReadSMUS bug or insufficient RAM."}; 
LOCAL char MsgForm[]        = { "(BAD_FORM) A malformed FORM SMUS." }; 
LOCAL char MsgShort[]       = { "(SHORT_CHUNK) A malformed FORM SMUS." }; 
LOCAL char MsgBad[]         = { "(BAD_IFF) A mangled IFF file." }; 
 
/* THESE MUST APPEAR IN RIGHT ORDER!! */ 
LOCAL char *IFFPMessages[-(int)LAST_ERROR+1] = { 
    /*IFF_OKAY*/  MsgOkay, 
    /*END_MARK*/  MsgEndMark, 
    /*IFF_DONE*/  MsgDone, 
    /*DOS_ERROR*/ MsgDos, 
    /*NOT_IFF*/   MsgNot, 
    /*NO_FILE*/   MsgNoFile, 
    /*CLIENT_ERROR*/ MsgClientError, 
    /*BAD_FORM*/  MsgForm, 
    /*SHORT_CHUNK*/  MsgShort, 
    /*BAD_IFF*/   MsgBad 
    }; 
 
typedef struct { 
   ClientFrame clientFrame; 
   UBYTE foundSHDR;
   UBYTE pad1;
   SScoreHeader *schd;
} SMUSFrame;


/* *********************************************************************** */ 
/* read a CHAR[] chunk */
IFFP ReadCHAR(context,destpt)
GroupContext *context;
char **destpt;
{
  IFFP iffp; 
  LONG szName;
  char *name;

  szName = ChunkMoreBytes(context); 
  name = (char *)calloc(szName,1); /* equiv. to malloc(szName) but init with zeros */

  if (name == NULL)
    iffp = CLIENT_ERROR;
  else {
    iffp = IFFReadBytes(context,(BYTE *)name,szName); 
    *destpt = name;
  }

  return(iffp);
}

/* *********************************************************************** */ 
/* read an instrument chunk */
IFFP ReadINS1 (context)
GroupContext *context;
{
  IFFP iffp;
  LONG szIns1;
  RefInstrument *ri;

  szIns1 = ChunkMoreBytes(context); 
  if (szIns1 > REFI_SIZE) return(BAD_FORM); /* chunk is too long */

  ri = AllocObject(RefInstrument);
  if (ri == NULL) return(CLIENT_ERROR);

  iffp = IFFReadBytes(context,(BYTE *)ri,szIns1); 
  CheckIFFP();

  if (ri->type == INS1_MIDI) {
    printf("WARNING:Instrument %s is midi\n",ri->name);
  }

  if (*(ri->name) == NULL) {
    printf("ERROR : instrument with no name\n");
    return(CLIENT_ERROR);
  }

  if (InstrList->its_num == MAX_INST) {
    printf("WARNING : Too many instruments for me, skipping %s\n",ri->name);
    return(iffp);
  }


  InstrList->its_instrs[InstrList->its_num] = ri;
  if (IRegs[ri->iRegister] != NO_INSTR) {
    printf("ERROR : instrument register already used\n");
    return(CLIENT_ERROR);
  }
  IRegs[ri->iRegister] = InstrList->its_num;
  InstrList->its_num +=1;
  return(iffp);
}
/* *********************************************************************** */ 
/* read a TRAK chunk */
IFFP ReadTRAK(context)
GroupContext *context;
{ 
    IFFP iffp; 
    LONG szTrack;
    struct Track *tr;

  if (The_Score->sc_ntc+1 > The_Score->sc_schd->ctTrack) return(BAD_FORM);
  /* more tracks than in header. Shouldn't be fatal */

  szTrack = ChunkMoreBytes(context); 
  
  tr = (struct Track *)calloc(szTrack+MIN_SZ_TRACK,1);

  if (tr == NULL) return(CLIENT_ERROR);

  tr->tr_len = szTrack/sizeof(SEvent);

  iffp = IFFReadBytes(context,(BYTE *)(&(tr->tr_fse[0])),szTrack); 

  The_Score->sc_tracks[The_Score->sc_ntc] = tr;
  The_Score->sc_ntc +=1;

  return(iffp);
}

/* *********************************************************************** */ 
IFFP GetFoSMUS(parent)
GroupContext *parent;
{ 
   IFFP iffp; 
   GroupContext formContext; 
   SMUSFrame smusFrame;          /* only used for non-clientFrame fields.*/ 
   SScoreHeader *schd;

   if (parent->subtype != ID_SMUS) 
      return(IFF_OKAY); /* just continue scaning the file */ 
 
   smusFrame = *(SMUSFrame *)parent->clientFrame; 
   iffp = OpenRGroup(parent,&formContext); 
   CheckIFFP(); 
 
   do switch (iffp = GetFChunkHdr(&formContext)) { 
      case ID_SHDR: { 
        smusFrame.foundSHDR = TRUE; 
        schd = AllocObject(SScoreHeader);
        if (schd != NULL) {
          iffp = GetSHDR(&formContext,schd);
          The_Score = (struct Score *)calloc(SZ_SCORE(schd->ctTrack));
          if (The_Score != NULL) {
            The_Score->sc_schd  = schd;
          }
          else 
            iffp = CLIENT_ERROR;
        }
        else
          iffp = CLIENT_ERROR;
       break; 
      } 
      case ID_NAME: { 
        if (!smusFrame.foundSHDR) 
            iffp = BAD_FORM;            /* Need an SHDR chunk first! */ 
        else iffp = ReadCHAR(&formContext,&(The_Score->sc_name)); 
        break; 
      } 
      case ID_AUTH: { 
        if (!smusFrame.foundSHDR) 
            iffp = BAD_FORM;            /* Need an SHDR chunk first! */ 
        else iffp = ReadCHAR(&formContext,&(The_Score->sc_auth)); 
        break; 
      } 
      case ID_ANNO: { 
        if (!smusFrame.foundSHDR) 
            iffp = BAD_FORM;            /* Need an SHDR chunk first! */ 
        else iffp = ReadCHAR(&formContext,&(The_Score->sc_anno)); 
        break; 
      } 
      case ID_Copyright: { 
        if (!smusFrame.foundSHDR) 
            iffp = BAD_FORM;            /* Need an SHDR chunk first! */ 
        else iffp = ReadCHAR(&formContext,&(The_Score->sc_copyr)); 
        break; 
      } 
      case ID_INS1: { 
        if (!smusFrame.foundSHDR) 
            iffp = BAD_FORM;            /* Need an SHDR chunk first! */ 
        else iffp = ReadINS1(&formContext); 
        break; 
      } 
      case ID_TRAK: { 
        if (!smusFrame.foundSHDR) 
            iffp = BAD_FORM;            /* Need an SHDR chunk first! */ 
        else iffp = ReadTRAK(&formContext); 
        break; 
      } 
      case END_MARK: { 
        if (!smusFrame.foundSHDR) 
            iffp = BAD_FORM; 
        else 
            iffp = IFF_DONE; 
        break; 
      } 
   } while (iffp >= IFF_OKAY);  /* loop if valid ID of ignored chunk or a 
                                   subroutine returned IFF_OKAY (no errors).*/ 
 
    if (iffp != IFF_DONE)  return(iffp); 
 
    /* If we get this far, there were no errors. */ 
    CloseRGroup(&formContext); 
    return(iffp); 
} 
 
/* *********************************************************************** */ 
IFFP ReadSMUS(file)
FILE *file;
{ 
   IFFP iffp = IFF_OKAY; 
   SMUSFrame sFrame;     /* Top level "client frame".*/ 

   sFrame.clientFrame.getList = SkipGroup; 
   sFrame.clientFrame.getProp = SkipGroup; 
   sFrame.clientFrame.getForm = GetFoSMUS; 
   sFrame.clientFrame.getCat  = ReadICat ; 
 
   sFrame.foundSHDR = FALSE; 
   sFrame.pad1      = 0; 

   InstrList = AllocStruct(Instruments);
   if (InstrList == NULL) {
     printf("out of memory for instruments list\n");
     return(CLIENT_ERROR);
   }
   InitLI(InstrList);

   iffp = ReadIFF(file,(ClientFrame *)&sFrame); 
 
   return(iffp); 
} 
 
 
/* *********************************************************************** */ 
BOOL getSMUS(filename)
char *filename;
{ 
  FILE *file;
  IFFP iffp = NO_FILE;
  char *temp;

  printf("Reading file '%s' \n",filename); 

  file = GOpen(filename,MODE_OLDFILE); 
  if (file!=NULL) {
     iffp = ReadSMUS(file); 
     GClose(file); 
  }

  if (iffp != IFF_DONE) {   
    printf("IFF error : %s for file %s\n",IFFPMessages[-iffp],filename); 
    return(FALSE);
  }

  if (The_Score->sc_name == NULL) {
    printf("WARNING : no NAME in score, using file name\n");
    temp = BaseName(filename);
    The_Score->sc_name = AllocString(strlen(temp));
    if (The_Score->sc_name == NULL) {
      printf("out of memory for score name\n");
      return(FALSE);
    }
    strcpy(The_Score->sc_name,temp);
  }

  if (The_Score->sc_ntc != The_Score->sc_schd->ctTrack) {
    printf("WARNING: less tracks than in header\n");
    printf("type 'y' to continue, anything else to abort :");
    if (GetRep()==0) return(FALSE);
    The_Score->sc_schd->ctTrack = The_Score->sc_ntc;
  }
  return(TRUE);
} 
/* *********************************************************************** */ 
